/*   BEM = Batch Equilibrium Model (calculates chemical equilibira)
     Copyright (C) 1996  J. Prenzel (jprenze@gwdg.de)
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */

#include <stdio.h>
#include <math.h>

#include "bem.h"

static int regoff();
static int wptmng();

/*------------------------------------------- slt0 --------------------*/
int slt0(){
  int i;
  for(i=0;i<Cl->nslt;i++){
    if(Cl->sl[i]->estslt > 0.){
      Cl->sl[i]->state = '+';
      Cl->sl[i]->wpt = Cl->sl[i]->estslt;
      Cl->sl[i]->cslt = Cl->sl[i]->cslt0;
    }
    else{ 
      Cl->sl[i]->state = '-';
    }
  }
  regoff();
  return 0;  
}

/*------------------------------------------- regoff -------------------*/
int regoff(){
  int i,ii;
  double d;

  for(i=0;i<Cl->nelt;i++){
    Cl->el[i]->ctoti = Cl->el[i]->ctot;
  }

  for(i=0;i<Cl->nslt;i++){
    if(Cl->sl[i]->state == '+' || Cl->sl[i]->state == '0'){
      if(Qind == 0){
	d = Cl->sl[i]->cslt - Cl->sl[i]->wpt;
	Cl->sl[i]->coff += d;
	Cl->sl[i]->cslt -= d;
      }
      for(ii=0;ii<Cl->sl[i]->nelt;ii++){
	Cl->sl[i]->el[ii]->ctoti -= 
          Cl->sl[i]->coff * Cl->sl[i]->stcf[ii] * Cl->rho; 
      }
      if(Cl->sl[i]->state == '0')Cl->sl[i]->state = '-'; 
    }
  }
  return 0;
}


/*------------------------------------------ sltmng --------------------*/
int sltmng(){
  int i,ii,rc,mng,m,k,ksimax=-999;
  double t,a,b,c,d,e,r,s,simax,esimax=1.;
  double lngam[MAXCHRG+1],oldwpt[MAXSLT];
  char oldstate[MAXSLT];


  for(i=0;i<Cl->nslt;i++){
    oldwpt[i]=Cl->sl[i]->wpt;
    oldstate[i]=Cl->sl[i]->state;
  }


  while(1){
    mng=0;
    wptmng();
    regoff();

    rc=findeq();
    Qind=0;

    Cl->it2cnt += Cl->it1cnt;
    if(rc==0 && Cl->it2cnt > Cl->it2max) rc=5;
    if(rc != 0){
      for(i=0;i<Cl->nslt;i++){
	Cl->sl[i]->wpt=oldwpt[i];
	Cl->sl[i]->state=oldstate[i];
      }
      regoff();
      return rc;
    } 
    else{
      d=Cl->ionstr;
      c=sqrt(d);
      a=-0.5*(c/(1.0+c) - 0.3*d)*M_LN10; 
      for(i=0;i<=MAXCHRG;i++){
	lngam[i]=a*i*i;
      }
      
      for(i=0;i<Cl->nslt;i++){
	if(Cl->sl[i]->state == '+'){
	  t = Cl->sl[i]->cslt + Cl->sl[i]->coff;
	  if(t < 0.){
	    Cl->sl[i]->wpt = t;
	    Cl->sl[i]->state = '0'; 
	    mng=1;
	    if(Cl->itrace>1)fprintf(Lgf.fil,"    exit %s\n",Cl->sl[i]->name);
	  }
	}
      }
      if(mng>0){regoff();}

      simax=0;
      for(i=0;i < Cl->nslt;i++){
	if(Cl->sl[i]->state == '-'){
	  spfun(i,c,&a,&b);
	  r=0;
	  s = 0.;  
	  for(ii=0; ii < Cl->sl[i]->nelt; ii++){
	    k = Cl->sl[i]->el[ii]->num;
	    b = Cl->sl[i]->stcf[ii];
	    r += b;
	    m=abs(Cl->el[k]->chrg);
	    s += b*(log(Cl->el[k]->csng) + lngam[m]);
	  }
/* 	  e = exp(s/r) - exp(a/r); */
	  e = exp(s/r);
	  s -= a;
	  if(s>simax){
	    simax=s;
	    ksimax=i;
	    esimax=e;
	  }
	}
      }
      if(simax > Cl->lsierr){
	mng=1;
	if(Cl->itrace>1)fprintf(Lgf.fil,"   enter %s\n",Cl->sl[ksimax]->name);
	Cl->sl[ksimax]->state = '+';
	k = Cl->nelt + Cl->niex + ksimax ;
	Cl->nus0.xarg[k] = log(esimax);
	if(Cl->sl[ksimax]->wptfix > 0.){
	  Cl->sl[ksimax]->wpt = Cl->sl[ksimax]->wptfix;
	}
	else{
	  Cl->sl[ksimax]->wpt = esimax * 10.; /* safety factor */
	}
      }
      for(i=0;i<Cl->nslt;i++){
	if(Cl->sl[i]->state == '0') Cl->sl[i]->state = '-';
      }
    }
    if(mng == 0){
      return 0;
    }
  }

  return 0;
}

/*----------------------------------------------- wptmng ----------------*/
int wptmng(){
  int i,ii,j,jj,k,pp;
  double a,w;

  if(Cl->step == 0){   /*  introduced 13.12.95  */
    return 0;
  }
  
  for(i=0;i<Cl->nslt;i++){
    if(Cl->sl[i]->state == '+'){
      if(Cl->sl[i]->wptfix > 0.){
	Cl->sl[i]->wpt = Cl->sl[i]->wptfix;
      }
      else{
	w=0;
	for(ii=0;ii<Cl->sl[i]->nelt;ii++){
	  k=Cl->sl[i]->el[ii]->num;
	  
	  if(k == 0 || k == 1) continue;
	  pp=0;
	  for(j=0;j<Cl->ngas;j++){
	    if(Cl->gs[j]->elnum == k && Cl->gs[j]->pfix > 0.) pp=1; 
	  }
	  if(pp == 1) continue;
	  
	  a=Cl->el[k]->csng;
	  for(j=0;j<Cl->ncpl;j++){
	    for(jj=0;jj<Cl->cp[j]->nelt;jj++){
	      if(Cl->cp[j]->el[jj]->num == k){
		a += Cl->cp[j]->ccpl * Cl->cp[j]->stcf[jj];
	      }
	    }
	  }
	  
	  a *= Cl->sl[i]->stcf[ii];
	  if(a>w) w=a;
	}
	Cl->sl[i]->wpt = w * 10 * Cl->theta / Cl->rho;
      }
    }
  }
  return 0;
}





